
from generallibrary import deco_cache, typeChecker, get_rows, initBases
from generalfile.optional_dependencies._extension import _Extension
from generalimport import fake_module_check

import pandas as pd
import csv


class Path_Spreadsheet:
    """ Spreadsheet methods for Path. """
    @property
    @deco_cache()
    def spreadsheet(self):
        """ Easy modify tsv and csv files. """
        fake_module_check(pd)
        return _Spreadsheet(self)


@initBases
class _Spreadsheet(_Extension):
    """ Contains all functionality to handle spreadsheet files.
        Such as CSV and TSV. """
    def write(self, df=None, overwrite=False):
        """ Can write a bunch of different DataFrames to a TSV file.
            Doesn't support advanced pandas functionality.
            Should work with: Keys, Index and Transposed (8 combinations)
            If DataFrame has both keys and index then cell A1 becomes NaN

            :param pandas.DataFrame df: Serializable by JSON
            :param overwrite: Whether to allow overwriting or not. """

        typeChecker(df, pd.DataFrame)

        with self.WriteContext(self.path, overwrite=overwrite) as write_path:
            if df.empty:
                self.path.write(overwrite=overwrite)
                return False, False

            useHeader = self._indexIsNamed(df.columns)
            useIndex = self._indexIsNamed(df.index)

            df.to_csv(write_path, sep="\t", header=useHeader, index=useIndex)

            return useHeader, useIndex

    def read(self, header=False, column=False, default=..., sep=None):
        """
        If any cell becomes NaN then the header and column parameters are overriden silently.

        Can read a bunch of different DataFrames to a TSV file.
        Doesn't support advanced pandas functionality.
        Should work with: Keys, Index, Transposed, Header, Column (32 combinations).
        DataFrame in file can have a NaN A1 cell.

        :param bool header: Use headers or not, overriden if any top left cell is NaN
        :param bool column: Use columns or not, overriden if any top left cell is NaN
        :param default:
        :param sep: Default to "\t"
        :rtype: pd.DataFrame
        """

        header = "infer" if header else None
        column = 0 if column else None

        with self.ReadContext(self.path) as read_path:
            try:
                df = self._read_helper(path=read_path, header=header, column=column, sep=sep)
            except pd.errors.EmptyDataError:
                return pd.DataFrame()
            except FileNotFoundError as e:
                if default is Ellipsis:
                    raise e
                else:
                    return default

            # Get rid of empty cell (Happens if file was written with header=True, column=True)
            headerFalseColumnFalse = pd.isna(df.iat[0, 0])
            headerFalseColumnTrue = pd.isna(df.index[0])
            headerTrueColumnFalse = str(df.columns[0]).startswith("Unnamed: ")
            if headerFalseColumnFalse or headerFalseColumnTrue or headerTrueColumnFalse:
                header = "infer"
                column = 0
                df = self._read_helper(path=read_path, header=header, column=column, sep=sep)
            else:
                # Get rid of name in index (Happens if file doesn't have an index and column=True)
                # Doesn't happen other way around for some reason, guess it's the internal order in pandas
                if df.index.name is not None:
                    if header is None and column == 0:
                        df.index.rename(None, inplace=True)
                    else:
                        header = None
                        column = None
                        df = self._read_helper(path=read_path, header=header, column=column, sep=sep)

            if not self._indexIsNamed(df.columns):
                df.columns = pd.RangeIndex(len(df.columns))
            if not self._indexIsNamed(df.index):
                df.index = pd.RangeIndex(len(df.index))

            return self._try_convert_dtypes(df)

    def append(self, obj):
        """
        Append an obj containing lists or dicts to the end of a TSV file.
        If a dict is given and there are iterables as values then the keys of the dict are the first value in each row.
        Otherwise keys in dicts are ignored.

        Identical append objects
         | [[1, 2, 3], [4, 5, 6]]
         | [{"a": 1, "b": 2, "c": 3}, {"d": 4, "e": 5, "f": 6}]
         | {1: {"b": 2, "c": 3}, 4: {"e": 5, "f": 6}}
         | {1: [2, 3], 4: [5, 6]}

        Todo: Support DataFrame and Series with spreadsheet.append()

        :param obj: Iterable (Optionally inside another iterable) or a value for a single cell
        """
        def _append_helper(stream):
            writer = csv.writer(stream, delimiter="\t", lineterminator="\n")
            for row in get_rows(obj):
                writer.writerow(row)

        with self.AppendContext(self.path) as append_path:
            return append_path.open_operation("a", _append_helper)


    def _indexIsNamed(self, index):
        """
        Simple version to see if a DataFrame index is named or not
        :param index: DataFrame's index (columns or index)
        """
        if not len(index) or str(index[0]) == "0" or str(index[0]) == "1":
            return False
        else:
            return True

    @staticmethod
    def _try_convert_dtypes(df):
        try:
            return df.convert_dtypes()
        except ValueError:
            return df

    def _read_helper(self, path, header, column, sep):
        if sep is None:
            sep = "\t"
        result = pd.read_csv(path, sep=sep, header=header, index_col=column)
        return self._try_convert_dtypes(result)

    def _append_helper(self, iterable_obj, key=None):
        """
        Takes an object and returns a list of rows to use for appending.

        :param iterable_obj: Iterable
        :param key: If iterableObj had a key to assigned it it's given here
        :return: A
        """
        row = [key] if key else []
        if isinstance(iterable_obj, (list, tuple)):
            row.extend(iterable_obj)
        elif isinstance(iterable_obj, dict):
            for _, value in sorted(iterable_obj.items()):
                row.append(value)
        return row












































class _MiddleGroundError(Exception):
    """Parent of all middle ground errors."""


class CaseSensitivityError(_MiddleGroundError):
    """Raised when an existing file matches a path but not the exact case."""


class InvalidCharacterError(_MiddleGroundError):
    """Raised when using an invalid character."""









from generallibrary import EnvVar

class _Path_Envs:
    """ Get information of path such as venv and package. """
    def is_venv(self):
        """ :param generalfile.Path self: """
        return (self / self.verInfo.venv_script_path / "activate").exists()

    def is_package(self):
        """ :param generalfile.Path self: """
        return bool(self.get_child(filt=lambda path: path.name() == "__pycache__"))

    def is_repo(self):
        """ :param generalfile.Path self: """
        return bool(self.get_child(filt=lambda x: x.name() in (".git", "metadata.json")))

    def get_parent_venv(self):
        """ Return first valid venv in parents or None

            :param generalfile.Path self:
            :rtype: generalfile.Path """
        return self.absolute().get_parent(depth=-1, include_self=True, traverse_excluded=True, filt=type(self).is_venv)

    def get_parent_package(self):
        """ Return last valid package in parents or None.

            :param generalfile.Path self:
            :rtype: generalfile.Path """
        return self.absolute().get_parent(depth=-1, index=-1, include_self=True, traverse_excluded=True, filt=type(self).is_package)

    def get_parent_repo(self):
        """ Return first valid repo in parents or None.

            :param generalfile.Path self:
            :rtype: generalfile.Path """
        return self.absolute().get_parent(depth=-1, include_self=True, traverse_excluded=True, filt=type(self).is_repo)

    @classmethod
    def get_active_venv_path(cls):
        """ :rtype: generalfile.Path """
        virtual_env = EnvVar("VIRTUAL_ENV", default=None).value
        if virtual_env is not None:
            venv_path = cls(virtual_env)
            assert venv_path.is_venv()
            return venv_path





































import pathlib

from generallibrary import VerInfo, TreeDiagram, Recycle, deco_cache

from generalfile.path_bases.path_lock import _Path_ContextManager
from generalfile.path_bases.path_operations import _Path_Operations
from generalfile.path_bases.path_strings import _Path_Strings
from generalfile.path_bases.path_envs import _Path_Envs
from generalfile.path_bases.path_scrub import _Path_Scrub
from generalfile.path_bases.path_diagram import _Path_Diagram


from generalfile.optional_dependencies.path_spreadsheet import Path_Spreadsheet
from generalfile.optional_dependencies.path_text import Path_Text
from generalfile.optional_dependencies.path_cfg import Path_Cfg
from generalfile.optional_dependencies.path_pickle import Path_Pickle


class Path(Recycle,
           _Path_ContextManager, _Path_Operations, _Path_Strings, _Path_Envs, _Path_Scrub, _Path_Diagram,
           Path_Spreadsheet, Path_Text, Path_Cfg, Path_Pickle):
    """ Immutable cross-platform Path.
        Built on pathlib and TreeDiagram.
        Implements rules to ensure cross-platform compatability.
        Adds useful methods.
        Todo: Binary extension. """
    verInfo = VerInfo()
    path_delimiter = verInfo.pathDelimiter
    Path = ...

    _recycle_keys = {"path": lambda path: Path.scrub(path)}
    _alternative_chars = {path_delimiter: "&#47;", ":": "&#58", ".": "&#46;"}


    def __init__(self, path=None):  # Don't have parent here because of Recycle
        self.path = self.scrub(path)

        self._path = pathlib.Path(self.path)
        self._latest_listdir = set()

    copy_node = NotImplemented  # Maybe something like this to disable certain methods

    def __str__(self):
        return getattr(self, "path", "<Path not loaded yet>")

    def __repr__(self):
        return f"<Path: '{self.path}'>"

    def __fspath__(self):
        return self.path

    def __format__(self, format_spec):
        return self.path.__format__(format_spec)

    def __truediv__(self, other):
        """ :rtype: generalfile.Path """
        return self.Path(self._path / str(other))

    @staticmethod
    @deco_cache()
    def _equal(str_pair):
        return

    def __eq__(self, other):
        if other is None:  # None in [Path()] was returning True without this
            return False

        if isinstance(other, Path):
            other = other.path
        else:
            other = self.scrub(other)
        return self.path == other

    def __hash__(self):
        return hash(self.path)

    def __contains__(self, item):
        return self.path.__contains__(item)

    def __dumps__(self):
        return str(self)

    @staticmethod
    def __loads__(path):
        return Path(path)

setattr(Path, "Path", Path)

















































from generallibrary import TreeDiagram
import os


class _Path_Diagram(TreeDiagram):
    def view_paths(self, spawn=True):
        """ :param generalfile.Path self: """
        self.view(spawn=spawn, custom_repr=lambda path: path.name())

    def spawn_parents(self):
        """ :param generalfile.Path self: """
        if not self.get_parent(spawn=False) and self.path and not self.is_root():
            try:
                index = self.path.rindex(self.path_delimiter) + 1
            except ValueError:
                index = 0
            self.set_parent(self.Path(path=self.path[:index]))

    def spawn_children(self):
        """ :param generalfile.Path self: """
        if self.is_folder():
            old_children = {path.name() for path in self.get_children(spawn=False)}

            try:
                new_children = set(os.listdir(self.path if self.path else "."))
            except PermissionError:
                new_children = set()

            for name in old_children - new_children:
                self.Path(path=self / name).set_parent(parent=None)

            for name in new_children - old_children:
                self.Path(path=self / name).set_parent(parent=self)
        else:
            self._children.clear()







from generallibrary import deco_cache, initBases
from generalfile.optional_dependencies._extension import _Extension

import dill


class Path_Pickle:
    """ Pickle methods for Path. """
    @property
    @deco_cache()
    def pickle(self):
        """ Easily write and read python objects with pickle. """
        return _Pickle(self)


@initBases
class _Pickle(_Extension):
    def write(self, obj=None, overwrite=False):
        """ Write to this path with a given object. """
        with self.WriteContext(self.path, overwrite=overwrite) as write_path:
            return write_path.open_operation("wb", lambda stream: dill.dump(obj=obj, file=stream), encoding=None)

    def read(self, default=...):
        """ Read from this path to get an object. """
        with self.ReadContext(self.path) as read_path:
            return read_path.open_operation("rb", lambda stream: dill.load(file=stream), encoding=None, no_file_default=default)













































from generallibrary import deco_cache, initBases, depth
from generalfile.optional_dependencies._extension import _Extension

import configparser
import json


class Path_Cfg:
    """ Cfg methods for Path. """
    @property
    @deco_cache()
    def cfg(self):
        """ Easily modify cfg files with a dictionary. """
        return _Cfg(self)


@initBases
class _Cfg(_Extension):
    _OBSCURE = "obscure_name_that_wont_match"
    _JSON_CAST = {
        "None": None,
        "True": True,
        "False": False,
    }
    _JSON_CAST_WRITE = {
        None: "None",
    }

    def _basic_write(self, dict_, overwrite):
        return self.path.text.write("\n".join(f"{key} = {value}" for key, value in dict_.items()), overwrite=overwrite)

    def write(self, dict_=None, overwrite=False):
        """ Write to this path with a given dictionary. """
        if depth(dict_) == 1:
            return self._basic_write(dict_=dict_, overwrite=overwrite)

        config = configparser.ConfigParser()
        # config = configparser.ConfigParser(converters={'list': lambda x: [i.strip() for i in x.split(',')]})

        cast_dict = self._write_cast_dict(dict_)

        config.read_dict(dictionary=cast_dict)
        with self.WriteContext(self.path, overwrite=overwrite) as write_path:
            return write_path.open_operation("w", lambda stream: config.write(stream))

    def _write_cast_dict(self, dict_):
        cast_dict = {}
        for section_name, section in dict_.items():
            cast_dict[section_name] = {key: self._write_json_cast(value=value) for key, value in section.items()}
        return cast_dict

    def _write_json_cast(self, value):
        try:
            if value in self._JSON_CAST_WRITE:
                return self._JSON_CAST_WRITE[value]
        except TypeError:
            pass

        return value

    def _read_json_cast(self, value):
        if value in self._JSON_CAST:
            return self._JSON_CAST[value]

        try:
            return json.loads(value.replace("'", '"'))
        except json.decoder.JSONDecodeError:
            return value

    def _read(self, config, path):
        string = path.text.read()
        try:
            config.read_string(string)
        except configparser.MissingSectionHeaderError:
            string = f"[{self._OBSCURE}]\n{string}"
            config.read_string(string)

    def _config_to_dict(self, config):
        dict_ = {section: {key: self._read_json_cast(value) for key, value in config.items(section)} for section in config.sections()}
        if self._OBSCURE in dict_:
            return dict_[self._OBSCURE]
        return dict_

    def read(self, default=...):
        """ Read from this path to get a dictionary. """
        config = configparser.ConfigParser()
        with self.ReadContext(self.path) as path:
            if not path.exists():
                if default is Ellipsis:
                    raise FileNotFoundError
                else:
                    return default
            self._read(config=config, path=path)
        return self._config_to_dict(config=config)

    def append(self, dict_):
        """ Update this cfg with a dictionary. """
        # append_path.cfg.write(dict_=append_path.cfg.read() | dict_, overwrite=True)  # 3.9
        self.write(dict_={**self.path.cfg.read(), **dict_}, overwrite=True)












































from copy import copy

from generallibrary import comma_and_or, Recycle, ObjInfo, deco_cache, AutoInitBases, DataClass, Log
from generalfile import Path

from itertools import chain


class _ConfigFile_ReadWrite:
    _CFG_HEADER_NAME = "config"

    def __init__(self):
        self._has_written = False
    
    def read_hook_pre(self): ...
    def read_hook_post(self): ...
    def write_hook_pre(self, dict_): ...
    def write_hook_post(self): ...

    def _read_JSON(self):
        """ :param ConfigFile self: """
        return self._path.read()

    def _read_CFG(self):
        """ :param ConfigFile self: """
        return self._path.cfg.read()[self._CFG_HEADER_NAME]

    @deco_cache()
    def _read_config(self):
        """ :param ConfigFile self: """
        if self.exists():
            self.read_hook_pre()

            read_methods = {"JSON": self._read_JSON, "CFG": self._read_CFG}
            read_method = read_methods[self._format]
            read_method_result = read_method()

            for key, value in read_method_result.items():
                if key in self.field_keys():
                    self.__dict__[key] = self._unserialize(key, value)  # Don't trigger __setattr__
                else:
                    Log().debug("Ignoring key value", key, value, "for", self._path)
            
            Log().debug("Reading config", self._path, read_method_result)

            self.read_hook_post()

    def _write_JSON(self, dict_):
        """ :param ConfigFile self: """
        self._path.write(content=dict_, overwrite=True, indent=4)

    def _write_CFG(self, dict_):
        """ :param ConfigFile self: """
        config_dict = {self._CFG_HEADER_NAME: dict_}
        self._path.cfg.write(dict_=config_dict, overwrite=True)

    def write_config(self):
        """ :param ConfigFile self: """
        dict_ = self.get_field_dict_serializable()
        self.write_hook_pre(dict_=dict_)
        write_methods = {"JSON": self._write_JSON, "CFG": self._write_CFG}
        write_method = write_methods[self._format]
        write_method(dict_=dict_)
        self._has_written = True
        Log().debug("Writing config", self._path, self.field_dict())
        self.write_hook_post()


class _ConfigFile_Serialize:
    @staticmethod
    def _has_serializers(value):
        return hasattr(value, "__dumps__") and hasattr(value, "__loads__")

    @classmethod
    def _serializable(cls, value):
        if value is not None and cls._has_serializers(value):
            return value.__dumps__()
        else:
            return copy(value)  # Prevent write_hook_pre from changing original mutable

    def _unserialize(self, key, value):
        if value is not None and key in self.get_custom_serializers():
            return self.get_custom_serializers()[key].__loads__(value)
        else:
            return value

    def _serializers_from_defaults(self):
        """ :param ConfigFile self: """
        return {key: type(value) for key, value in self.field_dict_defaults().items() if self._has_serializers(value=value)}

    def _serializers_from_annotations(self):
        """ :param ConfigFile self: """
        return {key: cls for key, cls in getattr(self, "__annotations__", {}).items() if self._has_serializers(value=cls)}

    @deco_cache()
    def get_custom_serializers(self):
        """ :param ConfigFile self: """
        combined = chain(
            self._serializers_from_defaults().items(),
            self._serializers_from_annotations().items(),
        )
        return {key: value for key, value in combined if key in self.field_keys()}

    def get_field_dict_serializable(self):
        """ :param ConfigFile self: """
        return {key: self._serializable(value) for key, value in self.field_dict().items()}




class ConfigFile(Recycle, DataClass, _ConfigFile_Serialize, _ConfigFile_ReadWrite, metaclass=AutoInitBases):
    """ Read config file when created.
        If value changes then write to file.
        Path must have support format suffix.
        Default value or annotation cls can define __dumps__ and __loads__.
        'name: str' will do nothing, must write 'name: str = None'
        Keys of dictionaries must be strings.

        Todo: Handle custom serializers within iterable for ConfigFile. """

    _supported_formats = {
        ".json": "JSON",
        ".cfg": "CFG",
    }

    _recycle_keys = {"path": lambda path: str(ConfigFile._scrub_path(path=path))}

    def __init__(self, path):
        self._path = self._scrub_path(path=path)
        self._format = self._supported_formats[self._path.suffix().lower()]

    def __repr__(self):
        return f"<{type(self).__name__} for '{self._path}'>"

    @classmethod
    def _scrub_path(cls, path):
        path = Path(path).absolute()
        assert path.suffix().lower() in cls._supported_formats, f"Path must end with {comma_and_or(*cls._supported_formats)}"
        return path

    def exists(self):
        if self._file_exists():
            return True
        return self._has_written

    @deco_cache()
    def _file_exists(self):
        return self._path.exists()

    @staticmethod
    def safe_equals(left, right):
        try:
            return left == right
        except:
            return False

    def __setattr__(self, key, value):
        prev_value = getattr(self, key, ...)
        super().__setattr__(key, value)
        if key in type(self).field_keys():
            # if prev_value != value:
            if not self.safe_equals(prev_value, value):  # StrictVersion struggled here, but should cover any scenario
                type(self).write_config(self)

    def __getattribute__(self, item):
        if not item.startswith("_") and item in type(self).field_keys():
            self._read_config()
        return super().__getattribute__(item)

    def halt_getattr(self, item):
        """ Get an item wihtout trigger __getattribute__"""
        if item in self.__dict__:
            return self.__dict__[item]
        else:
            return getattr(type(self), item)




from generalfile import Path
from generalfile.test.test_path import PathTest


class A:
    def __init__(self, x):
        self.x = x

class FileTest(PathTest):
    def test_pickle(self):
        Path("hi.txt").pickle.write(A(53))
        a = Path("hi.txt").pickle.read()
        self.assertEqual(53, a.x)

    def test_pickle_list(self):
        Path("hi.txt").pickle.write([A("hi"), A(53)])
        l = Path("hi.txt").pickle.read()
        self.assertEqual("hi", l[0].x)
        self.assertEqual(53, l[1].x)

    def test_read_empty(self):
        with self.assertRaises(FileNotFoundError):
            Path("hey").pickle.read()
        self.assertEqual(None, Path("hi").pickle.read(default=None))




from generallibrary import match, replace, deco_cache

from urllib.parse import quote


class _Path_Strings:
    """ String operations for Path. """
    def __getitem__(self, item):
        """ Get character from path string.

            :param generalfile.Path self: """
        return self.Path(self.path.__getitem__(item))

    @deco_cache()
    def to_alternative(self):
        """ Get path using alternative delimiter and alternative root for windows.

            :param generalfile.Path self:
            :rtype: generalfile.Path """
        return self.Path(replace(string=self.path, **self._alternative_chars))

    @deco_cache()
    def from_alternative(self):
        """ Get path from an alternative representation with or without leading lock dir.

            :param generalfile.Path self:
            :rtype: generalfile.Path """
        path = str(self.remove_start(self.get_lock_dir()))
        return self.Path(replace(string=path, reverse=True, **self._alternative_chars))

    def absolute(self):
        """ Get new Path as absolute.

            :param generalfile.Path self:
            :rtype: generalfile.Path """
        if self.is_absolute():
            return self
        else:
            return self.get_working_dir() / self

    def relative(self, base=None):
        """ Get new Path as relative, uses working dir if base is None.
            Returns self if not inside base.

            :param generalfile.Path self:
            :param base: Defaults to working dir. """
        if self.is_relative() and (base is None or not self.startswith(base)):
            return self
        else:
            if base is None:
                base = self.get_working_dir()
            try:
                return self.Path() if self == base else self.Path(self._path.relative_to(base))
            except ValueError:
                return None

    @deco_cache()
    def is_absolute(self):
        """ Get whether this Path is absolute.

            :param generalfile.Path self: """
        return self._path.is_absolute()

    @deco_cache()
    def is_relative(self):
        """ Get whether this Path is relative.

            :param generalfile.Path self: """
        return not self.is_absolute()

    @deco_cache()
    def mirror_path(self):
        """ Return mirror Path which currently points to same destination based on working dir.
            Absolute Path returns relative Path and vice versa.

            :param generalfile.Path self:
            :rtype: generalfile.Path """
        if self.is_absolute():
            return self.relative()
        else:
            return self.absolute()

    @deco_cache()
    def startswith(self, path):
        """ Get whether this Path starts with given string.

            :param generalfile.Path self:
            :param str or Path path:"""
        path = self.Path(path)
        return self.path.startswith(str(path))

    @deco_cache()
    def endswith(self, path):
        """ Get whether this Path ends with given string.

            :param generalfile.Path self:
            :param str or Path path:"""
        path = self.Path(path)
        return self.path.endswith(str(path))

    @deco_cache()
    def remove_start(self, path):
        """ Remove a string from the start of this Path if it exists.

            :param generalfile.Path self:
            :param str or Path path:"""
        path = self.Path(path)
        str_path = str(path)
        if not self.startswith(str_path):
            return self
        else:
            new_path = self.Path(self.path[len(str_path):])
            if str(new_path).startswith(path.path_delimiter):
                return new_path[1:]
            else:
                return new_path

    @deco_cache()
    def remove_end(self, path):
        """ Remove a string from the end of this Path if it exists.

            :param generalfile.Path self:
            :param str or Path path:"""
        path = self.Path(path)
        str_path = str(path)
        if not self.endswith(str_path):
            return self
        else:
            new_path = self.Path(self.path[:-len(str_path)])
            if str(new_path).endswith(path.path_delimiter):
                return new_path[:-1]
            else:
                return new_path

    def same_destination(self, path):
        """ See if two paths point to the same destination.

            :param generalfile.Path self:
            :param str or Path path:"""
        path = self.Path(path)
        return self.absolute() == path.absolute()

    @deco_cache()
    def parts(self):
        """ Split path using it's delimiter.
            With an absolute path the first index is an empty string on a posix system. <- Not sure about that anymore, might be /

            :param generalfile.Path self: """
        return self.path.split(self.path_delimiter)

    @deco_cache()
    def name(self):
        """ Get string name of Path which is stem + suffix, or entire path if root.

            :param generalfile.Path self:
            :rtype: str """
        return self.path if self.is_root() else self._path.name

    @deco_cache()
    def with_name(self, name):
        """ Get a new Path with new name which is stem + suffix.

            :param name: Name.
            :param generalfile.Path self:
            :rtype: generalfile.Path """
        return self.Path(self._path.with_name(str(name)))

    @deco_cache()
    def stem(self):
        """ Get stem which is name without last suffix.

            :param generalfile.Path self:
            :rtype: str """
        return self._path.stem

    @deco_cache()
    def with_stem(self, stem):
        """ Get a new Path with new stem which is name without last suffix.

            :param stem: New stem.
            :param generalfile.Path self:
            :rtype: generalfile.Path """
        return self.Path(self.with_name(f"{stem}{self.suffix()}"))

    @deco_cache()
    def true_stem(self):
        """ Get true stem which is name without any suffixes.

            :param generalfile.Path self:
            :rtype: str """
        return self._path.stem.split(".")[0]

    @deco_cache()
    def with_true_stem(self, true_stem):
        """ Get a new Path with new stem which is name without any suffixes.

            :param true_stem: New true stem.
            :param generalfile.Path self:
            :rtype: generalfile.Path """
        return self.Path(self.with_name(f"{true_stem}{''.join(self.suffixes())}"))

    @deco_cache()
    def suffix(self):
        """ Get suffix which is name without stem (e.g. .txt or .json).
            Empty string if missing.

            :param generalfile.Path self:
            :rtype: str """
        return self._path.suffix

    @deco_cache()
    def with_suffix(self, suffix, index=-1):
        """ Get a new Path with a new suffix at any index.
            Index is automatically clamped if it's outside index range.
            Set suffix to `None` to remove a suffix.

            :param generalfile.Path self:
            :param suffix: New suffix, can be `None`.
            :param index: Suffix index to alter.
            :rtype: generalfile.Path """

        suffixes = self.suffixes().copy()

        try:
            suffixes[index]
        except IndexError:
            if index >= len(suffixes):
                if not suffix:
                    if suffixes:
                        del suffixes[-1]
                else:
                    suffixes.append(suffix)
            else:
                if not suffix:
                    if suffixes:
                        del suffixes[0]
                else:
                    suffixes.insert(0, suffix)
        else:
            if not suffix:
                del suffixes[index]
            else:
                suffixes[index] = suffix

        return self.with_name(f"{self.true_stem()}{''.join(suffixes)}")

    @deco_cache()
    def suffixes(self):
        """ Get every suffix as a list.

            :param generalfile.Path self:
            :rtype: list[str] """
        return self._path.suffixes

    @deco_cache()
    def with_suffixes(self, *suffixes):
        """ Get a new Path with a new list of suffixes.

            :param str suffixes: New suffixes
            :param generalfile.Path self:
            :rtype: generalfile.Path """
        return self.Path(self.with_name(f"{self.true_stem()}{''.join(suffixes)}"))

    @deco_cache()
    def match(self, *patterns):
        """ Get whether this Path matches any given filter line.

            :param generalfile.Path self: """
        return match(self.path, *map(self._replace_delimiters, patterns))

    @deco_cache()
    def forward_slash(self):
        """ Return string path with forward slashes.

            :param generalfile.Path self: """
        return self.path.replace("\\", "/")

    @deco_cache()
    def encode(self):
        """ Return a URL encoded string from this Path.

            :param generalfile.Path self: """
        return quote(self.forward_slash())

    @deco_cache()
    def escaped(self):
        """ Return a path string with \\ instead of \.

            :param generalfile.Path self: """
        return self.path.replace("\\", "\\\\")




















from generalimport import generalimport
generalimport("pandas")

from generalfile.path import Path
from generalfile.configfile import ConfigFile
from generalfile.errors import CaseSensitivityError, InvalidCharacterError

Path.get_lock_dir().create_folder()
